from typing import Optional

from pydantic import BaseModel, Field


class Project(BaseModel):
    id: Optional[str] = Field(default=None)
    name: str
    api_key: Optional[str] = Field(alias="apiKey", default=None)
    api_secret: Optional[str] = Field(alias="apiSecret", default=None)

    def get_state_request_topic(self):
        return f"{self.id}/requests/states"

    def get_configuration_request_topic(self):
        return f"{self.id}/requests/configurations"

    class Config:
        populate_by_name = True



from datetime import datetime
from typing import List, Optional
import uuid as uuid_lib

from pydantic import BaseModel, Field, field_validator

from drax_sdk.model.dynamic import BaseValueObject


class ActivatorType:
    EVENT = "event"
    STATE = "state"
    SCHEDULED = "scheduled"


class Activator(BaseModel):
    type: str = Field(alias="type")
    selector: Optional[str] = Field(default=None, alias="selector")

    class Config:
        populate_by_name = True


class RuleVariable(BaseModel):
    name: str = Field(alias="name")
    label: str = Field(alias="label")
    value: str = Field(alias="value")

    class Config:
        populate_by_name = True


class ValuedDescriptor(BaseModel):
    variables: List[RuleVariable] = Field(alias="variables")

    class Config:
        populate_by_name = True


class ConditionDescriptor(ValuedDescriptor):
    uuid: str = Field(default_factory=lambda: str(uuid_lib.uuid4()), alias="uuid")
    type: str = Field(alias="type")

    class Config:
        populate_by_name = True


class ActionDescriptor(ValuedDescriptor):
    uuid: str = Field(default_factory=lambda: str(uuid_lib.uuid4()), alias="uuid")
    type: str = Field(alias="type")

    class Config:
        populate_by_name = True


class RuleDescriptor(BaseModel):
    id: Optional[str] = Field(default=None, alias="id")
    uuid: str = Field(default_factory=lambda: str(uuid_lib.uuid4()), alias="uuid")
    priority: int = Field(alias="priority")
    conditions: List[ConditionDescriptor] = Field(alias="conditions")
    actions: List[ActionDescriptor] = Field(alias="actions")
    is_or: bool = Field(alias="or")

    class Config:
        populate_by_name = True


class Automation(BaseModel):
    id: Optional[str] = Field(default=None, alias="id")
    code: str = Field(default="", alias="code")
    activator: Activator = Field(alias="activator")
    rules: List[RuleDescriptor] = Field(alias="rules")
    last_execution: Optional[datetime] = Field(alias="lastExecution", default=None)
    description: str = Field(alias="description")
    project_id: str = Field(alias="projectId")
    active: bool = Field(alias="active")
    last_activation: Optional[datetime] = Field(alias="lastActivation", default=None)
    idle_time_from_last_activation: int = Field(alias="idleTimeFromLastActivation", default=0)
    extras: Optional[BaseValueObject] = Field(alias="extras", default=None)
    modified: Optional[int] = Field(default=None, alias="modified")
    log_enabled: bool = Field(default=False, alias="logEnabled")

    class Config:
        populate_by_name = True



import os

import pandas as pd
from pydantic import BaseModel
from typing import List, Optional
import logging

DRAX_PUBLIC_KEY = "bb099d6dce1a953c7c5d2380815ee02ea191b39206000000ceefb3c222b480459556ce440379cef89db0ccfc04000000"


class NodeKey(BaseModel):
    node_id: str
    private_key: bytes


class KeyStore:
    drax_public_key: bytes = bytes.fromhex(DRAX_PUBLIC_KEY)
    keys: List[NodeKey] = []

    @classmethod
    def add(cls, node_id: str, private_key: bytes):
        """
        Add a key to the keystore
        :param node_id:
        :param private_key:
        :return:
        """
        key = NodeKey(node_id=node_id, private_key=private_key)
        cls.keys.append(key)

    @classmethod
    def add_hex(cls, node_id: str, private_key_hex: str):
        """
        Add a key to the keystore from a hex string
        :param node_id:
        :param private_key_hex:
        :return:
        """
        try:
            private_key = bytes.fromhex(private_key_hex)
            cls.add(node_id, private_key)
        except Exception as e:
            logging.warning(f"Hex key error for node {node_id}", e)

    @classmethod
    def get_private_key(cls, node_id: str) -> Optional[bytes]:
        """
        Get the private key for a node
        :param node_id:
        :return:
        """
        for key in cls.keys:
            if key.node_id == node_id:
                return key.private_key

        raise ValueError(f"Node {node_id} not found in keystore")

    @classmethod
    def save(cls):
        """
        save keys in a parquet file
        """

        df = pd.DataFrame(
            [(key.node_id, key.private_key.hex()) for key in cls.keys],
            columns=["node_id", "private_key"],
        )

        keystore_path = cls._get_keystore_path()

        df.to_parquet(keystore_path)

    @classmethod
    def load(cls):
        """
        load keys from a parquet file
        """

        keystore_path = cls._get_keystore_path()

        if not os.path.exists(keystore_path):
            return

        df = pd.read_parquet(keystore_path)

        for index, row in df.iterrows():
            cls.add_hex(row["node_id"], row["private_key"])

    @classmethod
    def _get_keystore_path(cls):
        project_root_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        keystore_path = os.path.join(project_root_path, "keystore.parquet")
        print(keystore_path)

        return keystore_path



from typing import List

import requests

from drax_sdk.model.dto import (
    Alarm,
    AlarmType,
    AlarmLogEntry,
    AcknowledgeAlarmRequest,
    ResolveAlarmRequest,
    FindAlarmLogEntriesRequest,
    FindAlarmTypesRequest,
    FindAlarmsRequest,
    PagedResult,
)


class AlarmsClient:
    service_url = ""
    api_key = ""
    api_secret = ""

    def __init__(self, service_url: str, api_key: str, api_secret: str):
        self.service_url = service_url
        self.api_key = api_key
        self.api_secret = api_secret

    def _get_headers(self):
        return {"drax-api-key": self.api_key, "drax-api-secret": self.api_secret}

    def get_alarm_collection_strategy_types(self) -> List[str]:
        """Get alarm collection strategy types"""
        response = requests.get(
            self.service_url + "/alarm-collection-strategy-types",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def get_alarm_scope_types(self) -> List[str]:
        """Get alarm scope types"""
        response = requests.get(
            self.service_url + "/alarm-scope-types",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def get_alarm_severity_types(self) -> List[str]:
        """Get alarm severity types"""
        response = requests.get(
            self.service_url + "/alarm-severity-types",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def get_alarm_status_types(self) -> List[str]:
        """Get alarm status types"""
        response = requests.get(
            self.service_url + "/alarm-status-types",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def get_alarm_type_statuses(self) -> List[str]:
        """Get alarm type statuses"""
        response = requests.get(
            self.service_url + "/alarm-type-statuses",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def acknowledge(self, acknowledge_request: AcknowledgeAlarmRequest) -> Alarm:
        """Acknowledge an alarm"""
        response = requests.post(
            self.service_url + "/alarm/acknowledge",
            json=acknowledge_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return Alarm.model_validate(response.json())

    def delete(self, project_id: str, alarm_id: str) -> None:
        """Delete an alarm"""
        response = requests.delete(
            self.service_url + f"/{project_id}/alarms/{alarm_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()

    def delete_alarm_type(self, alarm_type_id: str) -> None:
        """Delete an alarm type"""
        response = requests.delete(
            self.service_url + f"/alarm-types/{alarm_type_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()

    def find_alarm_log_entries(
        self, find_request: FindAlarmLogEntriesRequest
    ) -> PagedResult[AlarmLogEntry]:
        """Find alarm log entries"""
        response = requests.post(
            self.service_url + "/alarm-log-entries",
            json=find_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        data = response.json()
        # Extract from response envelope
        if "value" in data:
            data = data["value"]
        return PagedResult[AlarmLogEntry].model_validate(data)

    def find_alarm_types(self, find_request: FindAlarmTypesRequest) -> PagedResult[AlarmType]:
        """Find alarm types"""
        response = requests.post(
            self.service_url + "/alarm-types",
            json=find_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        data = response.json()
        # Extract from response envelope
        if "value" in data:
            data = data["value"]
        return PagedResult[AlarmType].model_validate(data)

    def find_alarms(self, find_request: FindAlarmsRequest) -> PagedResult[Alarm]:
        """Find alarms"""
        response = requests.post(
            self.service_url + "/alarms",
            json=find_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        data = response.json()
        # Extract from response envelope
        if "value" in data:
            data = data["value"]
        return PagedResult[Alarm].model_validate(data)

    def get_alarm(self, alarm_id: str) -> Alarm:
        """Get an alarm by id"""
        response = requests.get(
            self.service_url + f"/alarm/{alarm_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return Alarm.model_validate(response.json())

    def get_alarm_type(self, alarm_type_id: str) -> AlarmType:
        """Get an alarm type by id"""
        response = requests.get(
            self.service_url + f"/alarm-types/{alarm_type_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return AlarmType.model_validate(response.json())

    def resolve(self, resolve_request: ResolveAlarmRequest) -> Alarm:
        """Resolve an alarm"""
        response = requests.post(
            self.service_url + "/alarm/resolve",
            json=resolve_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return Alarm.model_validate(response.json())

    def save_alarm_type(self, alarm_type: AlarmType) -> AlarmType:
        """Save an alarm type"""
        response = requests.put(
            self.service_url + "/alarm-types",
            json=alarm_type.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return AlarmType.model_validate(response.json())



import base64
from typing import Optional

from pydantic import BaseModel, field_validator, field_serializer


class BytesBase64Model(BaseModel):

    @field_validator("*", mode="before")
    def decode_base64_bytes(cls, v, info):
        if isinstance(v, str):
            field_name = info.field_name
            model_field = cls.model_fields.get(field_name)
            if model_field is None:
                return v

            # check if is bytes or optional bytes
            if (
                model_field.annotation == bytes
                or model_field.annotation == Optional[bytes]
            ):
                return base64.b64decode(v)
        return v

    @field_serializer("*")
    def serialize_field(self, v, field):
        if isinstance(v, bytes):
            return base64.b64encode(v).decode("utf-8")
        return v



import json
import logging
import threading
import time
import uuid
from abc import ABC, abstractmethod
from typing import Callable, List, TypeVar, Generic

import pika
from pika import SelectConnection
from pika.channel import Channel

from drax_sdk.model.config import DraxConfigParams
from drax_sdk.model.dto import ConfigurationResponse, StateResponse
from drax_sdk.model.event import Event
from drax_sdk.model.node import Configuration, State
from drax_sdk.utils.codec import decode_configuration
from drax_sdk.utils.keystore import KeyStore

logger = logging.getLogger(__name__)


T = TypeVar("T")


class Listener(ABC, Generic[T]):
    id: str = uuid.uuid4().hex
    routing_key: str
    queue: str | None = None
    callback: Callable[[T], None]
    exchange: str = "amq.topic"
    auto_ack: bool = True
    node_id: str | None = None

    _channel: Channel

    def __init__(
        self,
        exchange: str,
        routing_key: str,
        callback: Callable[[T], None],
        queue: str = "",
        auto_ack: bool = True,
        node_id: str = None,
    ):
        self.exchange = exchange
        self.routing_key = routing_key
        self.callback = callback
        self.queue = queue
        self.auto_ack = auto_ack
        self.node_id = node_id

    def on_queue_declared(self, method_frame):
        self.queue = method_frame.method.queue
        self._channel.queue_bind(
            exchange=self.exchange,
            queue=self.queue,
            routing_key=self.routing_key,
            callback=self.on_queue_bound,
        )

    def on_queue_bound(self, method_frame):
        self._channel.basic_consume(
            queue=self.queue,
            on_message_callback=self.on_message,
            auto_ack=self.auto_ack,
        )

    @abstractmethod
    def on_message(self, ch, method, properties, body):
        raise NotImplementedError

    def start_consuming(self, channel: Channel):
        self._channel = channel

        if self._channel is None:
            return

        self._channel.queue_declare(
            queue=self.queue,
            durable=False,
            exclusive=True,
            auto_delete=True,
            callback=self.on_queue_declared,
        )

    def stop_consuming(self):
        if (
            self._channel is not None
            and self.queue is not None
            and self._channel.is_open
        ):
            self._channel.queue_unbind(
                queue=self.queue,
                exchange=self.exchange,
                routing_key=self.routing_key,
            )
            self._channel.basic_cancel(self.queue)


class ConfigurationListener(Listener[Configuration]):

    def __init__(
        self,
        routing_key: str,
        callback: Callable[[Configuration], None],
        node_id: str | None = None,
    ):
        super().__init__(
            routing_key=routing_key,
            callback=callback,
            exchange="amq.topic",
            queue="",
            node_id=node_id,
        )

    def on_message(self, ch, method, properties, body):
        try:
            logger.debug("Received configuration: %r", body)
            json_data = json.loads(body.decode("utf-8"))
            configuration_response = ConfigurationResponse.model_validate(json_data)

            if self.node_id is not None:
                if configuration_response.node_id != self.node_id:
                    logger.debug(
                        "Configuration for node %s not for listener %s. Ignoring...",
                        configuration_response.node_id,
                        self.id,
                    )
                    return

            if configuration_response.cryptography_disabled:
                configuration_map = json.loads(
                    configuration_response.configuration.decode()
                )
            else:
                node_private_key = KeyStore.get_private_key(
                    configuration_response.node_id
                )

                configuration_map = decode_configuration(
                    node_private_key, configuration_response.configuration
                )

            configuration = Configuration.from_dict(configuration_map)
            configuration.timestamp = configuration_response.timestamp
            configuration.node_id = configuration_response.node_id

            self.callback(configuration)
        except Exception as e:
            logger.error("Error processing configuration: %s", e)


class StateListener(Listener[State]):
    def __init__(
        self,
        routing_key: str,
        callback: Callable[[State], None],
        node_id: str | None = None,
    ):
        super().__init__(
            routing_key=routing_key,
            callback=callback,
            exchange="amq.topic",
            queue="",
            node_id=node_id,
        )

    def on_message(self, ch, method, properties, body):
        try:
            logger.debug("Received state: %r", body)
            json_data = json.loads(body.decode("utf-8"))
            state_response = StateResponse.model_validate(json_data)

            if self.node_id is not None:
                if state_response.node_id != self.node_id:
                    logger.debug(
                        "State for node %s not for listener %s. Ignoring...",
                        state_response.node_id,
                        self.id,
                    )
                    return

            state = State.from_dict(state_response.state)
            state.node_id = state_response.node_id
            state.timestamp = state_response.timestamp

            self.callback(state)
        except Exception as e:
            logger.error("Error processing state: %s", e)


class EventListener(Listener[Event]):
    def __init__(
        self,
        routing_key: str,
        callback: Callable[[Event], None],
        node_id: str | None = None,
    ):
        super().__init__(
            routing_key=routing_key,
            callback=callback,
            exchange="amq.topic",
            queue="",
            node_id=node_id,
        )

    def on_message(self, ch, method, properties, body):
        try:
            logger.debug("Received event: %r", body)
            json_data = json.loads(body.decode("utf-8"))
            event = Event.model_validate(json_data)

            if event:
                self.callback(event)
        except Exception as e:
            logger.error("Error processing event: %s", e)


class DraxAmqpBroker:

    _thread: threading.Thread | None
    _configuration: DraxConfigParams
    _connection: SelectConnection | None
    _channel: SelectConnection | None
    _listeners: List[Listener]
    _running = False

    def __init__(self, configuration: DraxConfigParams):
        self._thread = None
        self._configuration = configuration
        self._connection = None
        self._channel = None
        self._listeners = []

    def _connect(self):
        logger.debug(
            f"Connecting to AMQP broker at host {self._configuration.broker_host}, vhost {self._configuration.broker_vhost}"
        )
        params = pika.ConnectionParameters(
            host=self._configuration.broker_host,
            virtual_host=self._configuration.broker_vhost,
            credentials=pika.PlainCredentials(
                self._configuration.api_key, self._configuration.api_secret
            ),
        )
        self._connection = pika.SelectConnection(
            params,
            on_open_callback=self.on_connection_open,
            on_close_callback=self.on_connection_closed,
            on_open_error_callback=self.on_connection_error,
        )

        self._connection.ioloop.start()

    def on_connection_open(self, unused_connection):
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        self._channel = channel
        for listener in self._listeners:
            listener.start_consuming(channel)

    def on_connection_error(self, connection, error):
        logger.error("Connection error: %s", repr(error))
        self._connection.ioloop.stop()

        if self._thread and self._thread.is_alive() and self._running:
            logger.debug("Reconnecting...")
            time.sleep(5)
            self._connect()

    def add_configuration_listener(
        self,
        topic: str,
        cb: Callable[[Configuration], None],
        project_id: str = None,
        node_id: str = None,
    ):
        project_id = project_id or self._configuration.project_id
        project_topic = f"{project_id}/{topic}".replace("/", ".")

        listener = ConfigurationListener(
            routing_key=project_topic, callback=cb, node_id=node_id
        )
        self._listeners.append(listener)

        logger.debug(
            "Registered configuration listener %s for topic: %s",
            listener.id,
            project_topic,
        )

        listener.start_consuming(self._channel)

    def add_state_listener(
        self,
        topic: str,
        cb: Callable[[State], None],
        project_id: str = None,
        node_id: str = None,
    ):
        project_id = project_id or self._configuration.project_id
        project_topic = f"{project_id}/{topic}".replace("/", ".")

        listener = StateListener(
            routing_key=project_topic, callback=cb, node_id=node_id
        )
        self._listeners.append(listener)

        logger.debug(
            "Registered state listener %s for topic: %s",
            listener.id,
            project_topic,
        )

        listener.start_consuming(self._channel)

    def add_event_listener(self, cb: Callable[[Event], None], project_id: str = None):
        project_id = project_id or self._configuration.project_id
        project_topic = f"{project_id}/events".replace("/", ".")

        listener = EventListener(routing_key=project_topic, callback=cb)
        self._listeners.append(listener)

        logger.debug(
            "Registered events listener %s for topic: %s",
            listener.id,
            project_topic,
        )

        listener.start_consuming(self._channel)

    def start(self):
        logger.debug("Starting broker...")
        self._running = True
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def stop(self):
        logger.debug("Stopping broker...")
        self._running = False
        if self._connection is not None:
            for listener in self._listeners:
                listener.stop_consuming()

            self._channel.close()
            self._connection.close()
            self._connection.ioloop.stop()

            logger.debug("Broker stopped")

    def on_connection_closed(self, connection, reason):
        logger.debug("Connection closed: %s", reason)

        for listener in self._listeners:
            listener.stop_consuming()

        self._connection.ioloop.stop()
        if self._thread and self._thread.is_alive() and self._running:
            logger.debug("Reconnecting...")
            time.sleep(5)  # delay before attempting to reconnect
            self._connect()
            self._connection.ioloop.start()

    def _run(self):
        self._connect()



from typing import List

import requests

from drax_sdk.model.dto import (
    HandshakeRequest,
    StateRequest,
    ConfigurationRequest,
    FindNodeByIdsRequest,
    HandshakeResponse,
    PagedResult,
    InstalledNode,
    StateResponse,
    FlatConfigurationResponse,
    InstallRequest,
    PrepareRequest,
)
from drax_sdk.model.event import Event
from drax_sdk.model.node import Node, NodeType
from drax_sdk.model.project import Project


class DraxCoreClient:
    DEFAULT_PAGE = 1
    DEFAULT_SIZE = 10

    service_url = ""
    api_key = ""
    api_secret = ""

    def __init__(self, service_url: str, api_key: str, api_secret: str):
        self.service_url = service_url
        self.api_key = api_key
        self.api_secret = api_secret

    def _get_headers(self):
        return {"drax-api-key": self.api_key, "drax-api-secret": self.api_secret}

    def handshake(self, request: HandshakeRequest) -> HandshakeResponse:
        payload = request.model_dump(by_alias=True)
        response = requests.post(
            self.service_url + "/handshake",
            json=payload,
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()

        return HandshakeResponse.model_validate(response.json())

    def register_node_type(self, node_type: NodeType) -> NodeType:
        response = requests.post(
            self.service_url + "/nodeTypes",
            headers=self._get_headers(),
            json=node_type.model_dump(by_alias=True),
        )
        response.raise_for_status()

        return NodeType.model_validate(response.json())

    def update_node_type(self, node_type: NodeType):
        response = requests.post(
            self.service_url + f"/nodeTypes/{node_type.id}",
            headers=self._get_headers(),
            json=node_type.model_dump(by_alias=True),
        )
        response.raise_for_status()

    def get_node_type_by_id(self, node_type_id: str) -> NodeType:
        response = requests.get(
            self.service_url + f"/nodeTypes/{node_type_id}", headers=self._get_headers()
        )
        response.raise_for_status()

        return NodeType.model_validate(response.json())

    def unregister_node_type(self, node_type_id: str):
        response = requests.delete(
            self.service_url + f"/nodeTypes/{node_type_id}", headers=self._get_headers()
        )
        response.raise_for_status()

    def list_node_types(
            self,
            project_id: str = None,
            keyword: str = None,
            page: int = DEFAULT_PAGE,
            size: int = DEFAULT_SIZE
    ) -> PagedResult[NodeType]:
        params = {
            "page": page,
            "size": size,
        }
        if project_id:
            params["projectId"] = project_id
        if keyword:
            params["keyword"] = keyword
        response = requests.get(
            self.service_url + "/nodeTypes",
            headers=self._get_headers(),
            params=params,
        )
        response.raise_for_status()

        return PagedResult[NodeType].model_validate(response.json())

    def list_projects(
            self,
            page: int = DEFAULT_PAGE,
            size: int = DEFAULT_SIZE
    ) -> PagedResult[Project]:
        response = requests.get(
            self.service_url + "/projects",
            headers=self._get_headers(),
            params={
                "page": page,
                "size": size,
            },
        )
        response.raise_for_status()

        return PagedResult[Project].model_validate(response.json())

    def register_project(self, project: Project) -> Project:
        response = requests.post(
            self.service_url + "/projects",
            headers=self._get_headers(),
            json=project.model_dump(by_alias=True),
        )
        response.raise_for_status()

        return Project.model_validate(response.json())

    def update_project(self, project: Project):
        response = requests.post(
            self.service_url + f'/projects/{project["id"]}',
            headers=self._get_headers(),
            json=project.model_dump(by_alias=True),
        )
        response.raise_for_status()

    def get_project_by_id(self, project_id: str) -> Project:
        response = requests.get(
            self.service_url + f"/projects/{project_id}", headers=self._get_headers()
        )
        response.raise_for_status()

        return Project.model_validate(response.json())

    def unregister_project(self, project_id: str):
        response = requests.delete(
            self.service_url + f"/projects/{project_id}", headers=self._get_headers()
        )
        response.raise_for_status()

    def get_my_project(self) -> Project:
        response = requests.get(
            self.service_url + "/projects/my", headers=self._get_headers()
        )
        response.raise_for_status()

        return Project.model_validate(response.json())

    def get_next_project_id(self, project_id: str) -> str:
        response = requests.get(
            self.service_url + f"/projectsIds/next/{project_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.text

    def prepare_node(self, request: PrepareRequest) -> InstalledNode:
        response = requests.post(
            self.service_url + "/prepared-nodes",
            headers=self._get_headers(),
            json=request.model_dump(by_alias=True),
        )
        response.raise_for_status()

        return InstalledNode.model_validate(response.json())

    def install_node(self, request: InstallRequest) -> InstalledNode:
        response = requests.post(
            self.service_url + "/nodes",
            headers=self._get_headers(),
            json=request.model_dump(by_alias=True),
        )
        response.raise_for_status()

        return InstalledNode.model_validate(response.json())

    def update_node(self, node: Node):
        response = requests.post(
            self.service_url + f"/nodes/{node.id}",
            headers=self._get_headers(),
            json=node.model_dump(by_alias=True),
        )

        response.raise_for_status()

    def uninstall_node(self, node_id: str):
        response = requests.delete(
            self.service_url + f"/nodes/{node_id}", headers=self._get_headers()
        )
        response.raise_for_status()

    def get_node_by_id(self, node_id: str) -> Node:
        response = requests.get(
            self.service_url + f"/nodes/{node_id}", headers=self._get_headers()
        )
        response.raise_for_status()

        return Node.model_validate(response.json())

    def list_nodes(
            self,
            project_id: str = None,
            keyword: str = None,
            page: int = DEFAULT_PAGE,
            size: int = DEFAULT_SIZE
    ) -> PagedResult[Node]:
        params = {
            "page": page,
            "size": size,
        }
        if project_id:
            params["projectId"] = project_id
        if keyword:
            params["keyword"] = keyword
        response = requests.get(
            self.service_url + "/nodes", params=params, headers=self._get_headers()
        )
        response.raise_for_status()

        return PagedResult[Node].model_validate(response.json())

    def get_state(self, node_id: str) -> StateResponse:
        url = self.service_url + f"/nodes/{node_id}/state"
        response = requests.get(url, headers=self._get_headers())

        response.raise_for_status()

        return StateResponse.model_validate(response.json())

    def list_states(
        self,
        node_id: str,
        project_id: str = None,
        from_time: int = None,
        to_time: int = None,
        page: int = DEFAULT_PAGE,
        size: int = DEFAULT_SIZE,
    ) -> PagedResult[StateResponse]:
        params = {
            "page": page,
            "size": size,
        }
        if project_id:
            params["projectId"] = project_id
        if from_time:
            params["from"] = from_time
        if to_time:
            params["to"] = to_time
        response = requests.get(
            self.service_url + f"/nodes/{node_id}/states",
            params=params,
            headers=self._get_headers(),
        )
        response.raise_for_status()

        return PagedResult[StateResponse].model_validate(response.json())

    def list_configurations(
        self,
        node_id: str,
        from_time: int = None,
        to_time: int = None,
        page: int = DEFAULT_PAGE,
        size: int = DEFAULT_SIZE,
    ) -> PagedResult[FlatConfigurationResponse]:
        params = {
            "from": from_time,
            "to": to_time,
            "page": page,
            "size": size,
        }
        response = requests.get(
            self.service_url + f"/nodes/{node_id}/configurations",
            params=params,
            headers=self._get_headers(),
        )
        response.raise_for_status()

        return PagedResult[FlatConfigurationResponse].model_validate(response.json())

    def list_nodes_states(
        self,
        project_id: str | None,
        find_node_by_ids_request: FindNodeByIdsRequest,
        from_time: int = None,
        to_time: int = None,
        page: int = DEFAULT_PAGE,
        size: int = DEFAULT_SIZE,
    ) -> PagedResult[StateResponse]:
        params = {
            "page": page,
            "size": size,
        }
        if project_id:
            params["projectId"] = project_id
        if from_time:
            params["from"] = from_time
        if to_time:
            params["to"] = to_time
        response = requests.post(
            self.service_url + "/states/find-by-node-ids",
            params=params,
            headers=self._get_headers(),
            json=find_node_by_ids_request.model_dump(by_alias=True),
        )
        response.raise_for_status()

        return PagedResult[StateResponse].model_validate(response.json())

    def set_state(self, node_id, request: StateRequest):
        response = requests.post(
            self.service_url + f"/nodes/{node_id}/states",
            headers=self._get_headers(),
            json=request.model_dump(by_alias=True),
        )
        response.raise_for_status()

    def set_configuration(self, node_id, request: ConfigurationRequest):
        response = requests.post(
            self.service_url + f"/nodes/{node_id}/configurations",
            headers=self._get_headers(),
            json=request.model_dump(by_alias=True),
        )
        response.raise_for_status()

    def request_state_update(self, node_id: str):
        response = requests.post(
            self.service_url + f"/nodes/{node_id}/states/update",
            headers=self._get_headers(),
        )
        response.raise_for_status()

    def invoke(self, event: Event):
        response = requests.post(
            self.service_url + "/events/invoke",
            headers=self._get_headers(),
            json=event.model_dump(by_alias=True),
        )
        response.raise_for_status()

    def get_nodes_by_ids(self, request: FindNodeByIdsRequest) -> List[Node]:
        response = requests.post(
            self.service_url + "/nodes/find-by-ids",
            headers=self._get_headers(),
            json=request.model_dump(by_alias=True),
        )
        response.raise_for_status()

        return [Node.model_validate(node) for node in response.json()]
